/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "lmsg/Buffer.hh"
#include "lmsg/BufferPool.hh"
#include "lmsg/Message.hh"
#include "lmsg/ErrorList.hh"
#include "lmsg/TCPSocket.hh"
#include "lmsg/SocketPool.hh"
#include "lmsg/TransportTCP.hh"
#include "lmsg/TransInput.hh"
#include "lmsg/TransOutput.hh"
#include <stdio.h>
#include <iostream>

using namespace lmsg;
using namespace std;

//======================================  Constructor
TransportTCP::TransportTCP(void) 
  : mCSocket(0)
{
}

//======================================  Destructor
TransportTCP::~TransportTCP(void) {
    close();
}

//======================================  Is socket connected?
bool
TransportTCP::isConnected(void) const {
    if (!isOpen()) return false;
    if (mCSocket) return true;
    return mSocket->isConnected();
}

//======================================  Close the socket
void
TransportTCP::close(void) {
    if (mCSocket) {
        disconnect();
    }

    if (mSocket) {
        disconnect();
	mSocket->async(false);
        mSocket->Return();
	mSocket = 0;
    }
}

//======================================  Connect to a partner.
error_type 
TransportTCP::connect(const MsgAddr& addr) {
    if (!isOpen()) return NotOpen;
    error_type rc = mSocket->connect(addr);
    if (rc == SystemError && mDebug) {
        ::perror("Error in connect:");
    } else if (rc && mDebug) {
        cerr << "Error in connect: " << rc << endl;
    } else if (mDebug > 1) {
        cout << "Made connection to " << addr << endl;
    }
    return rc;
}

//======================================  Break connection.
error_type 
TransportTCP::disconnect(void) {
    error_type rc;
    if (!isOpen()) return NotOpen;
    if (isClient()) {
        rc = mSocket->disconnect();
    } else if (mCSocket) {
        rc = mCSocket->disconnect();
        mCSocket->Return();
	mCSocket = 0;
    }
    return OK;
}

//======================================  Open the socket
error_type 
TransportTCP::open(mask_type flags, const MsgAddr* addr) {
    mFlags  = flags;
    mSocket = new TCPSocket(addr);
    if (!isOpen()) return NotOpen;
    mSocket->setDebug(mDebug);
    if (is_async()) mSocket->async(true);
    mSocket->setPool(mBuffPool);
    if (!isClient()) {
	dynamic_cast<TCPSocket*>(mSocket)->listen(LISTEN_QUEUE_DEPTH);
    }
    error_type rc = mSocket->setSendBuf(1<<20);
    rc = mSocket->setRecvBuf(1<<20);
    return rc;
}

//======================================  Kill the socket?
void 
TransportTCP::Purge(void) {
    close();
}

//======================================  Receive a message
error_type 
TransportTCP::receive(MsgHeader& hdr, char* Data, size_type maxlen, 
			    wtime_type timeout) {
    Buffer* buf;
    index_type rc = receive(&buf, timeout);
    if (rc) return rc;
    hdr = buf->getHeader();
    size_type len = hdr.getMsgLength();
    if (len > maxlen) rc = SizeError;
    else              memcpy(Data, buf->getDataAddr(), len);
    buf->Return();
    return rc;
}

//======================================  Receive a message
error_type 
TransportTCP::receive(MsgHeader& hdr, Message& msg, wtime_type timeout) {
    Buffer* buf;
    index_type rc = receive(&buf, timeout);
    if (rc) return rc;
    hdr = buf->getHeader();
    if (hdr.getMsgType() != msg.getType()) {
        if (getDebug()) {
	    cout << "Message type: " << hdr.getMsgType() << " not expected (" 
		 << msg.getType() << ")" << endl;
	}
        rc = Invalid;
    } else {
        TransInput in(*buf);
	try {
	    msg.setData(in);
	} catch (exception& e) {
	    if (getDebug()) {
	        cout << "Exception: " << e.what() 
		     << " in TransportTCP::receive" << endl;
	    }
	    rc = SizeError;
	}
    }
    buf->Return();
    return rc;
}

//======================================  Receive a message
error_type 
TransportTCP::receive(Buffer** Data, wtime_type timeout) {
    MsgHeader hdr;
    error_type rc;

    //---------------------------------- Get the message header
    if (!isOpen()) return NotOpen;
    TCPSocket* sock = 0;
    if (isClient()) sock = dynamic_cast<TCPSocket*>(mSocket);
    else            sock = dynamic_cast<TCPSocket*>(mCSocket);
    rc = sock->read(reinterpret_cast<char*>(&hdr), sizeof(hdr));
    if (rc) return rc;
    hdr.Import();
    hdr.getSource() = sock->getPartner();
    MsgAddr* s = &hdr.getSource();
    if (s->getIPAddr() == 0) s->setIPAddr(sock->getPartner().getIPAddr());
    if (getDebug() > 1) hdr.Dump(cout);

    //---------------------------------- Get an appropriate buffer.
    size_type msgLen = hdr.getMsgLength();
    if (mBuffPool && mBuffPool->getLength() >= msgLen+sizeof(MsgHeader)) {
        *Data = mBuffPool->getBuffer();
    } else {
        *Data = new Buffer(msgLen+sizeof(MsgHeader));
    }
    if (!*Data) return NoBuffer;
    (*Data)->getHeader() = hdr;
    rc =  sock->read((*Data)->getDataAddr(), msgLen);
    if (rc) {
        (*Data)->Return();
        *Data = 0;
        return rc;
    }
    (*Data)->setDataLength(msgLen); //set the length inclusive of header.
    return rc;
}

//======================================  Send a message
error_type 
TransportTCP::send(const MsgHeader& hdr, const char* b) {
    size_type msgLen = hdr.getMsgLength();
    Buffer* buf;
    if (mBuffPool && mBuffPool->getDataLength() > msgLen) {
        buf = mBuffPool->getBuffer();
    } else {
        buf = new Buffer(msgLen);
    }
    if (!buf) return NoBuffer;
    buf->getHeader() = hdr;
    memcpy(buf->getDataAddr(), b, msgLen);
    buf->setDataLength(msgLen);
    return send(buf);
}

//======================================  Send a message
error_type 
TransportTCP::send(const MsgHeader& hdr, const Message& msg) {
    if (!hdr.getDest()) return BadAddress;

    //----------------------------------  Get an appropriate buffer.
    Buffer* buf;
    size_type msgLen = msg.size() + sizeof(hdr);
    if (mBuffPool && mBuffPool->getDataLength() >= msgLen) {
        buf = mBuffPool->getBuffer();
    } else {
        buf = new Buffer(msgLen);
    }
    if (!buf) return NoBuffer;
    buf->getHeader() = hdr;

    //----------------------------------  Fill the buffer.
    try {
	TransOutput out(*buf);
        msg.getData(out);
	index_type len = out.getNBytes();
	buf->setDataLength(len);
	buf->getHeader().setMsgLength(len);
	buf->getHeader().setMsgType(msg.getType());
    } catch (exception& e) {
        if (getDebug()) {
	    cout << "Exception: " << e.what() << " in TransportTCP::send" 
		 << endl;
	    cout << "Output buffer length: " << buf->getSize() << endl;
	}
	buf->Return();
	return SizeError;
    }
    return send(buf);
}

//======================================  Send message contained in a buffer
error_type 
TransportTCP::send(Buffer* b) {
    if (!isOpen()) return NotOpen;
    b->getHeader().setSource(getAddr());
    if (getDebug() > 1) b->getHeader().Dump(cout);
    b->getHeader().Export();
    error_type rc;
    if (isClient()) rc = mSocket->send(b->getHeader().getDest(), b);
    else            rc = mCSocket->send(b->getHeader().getDest(), b);
    if (rc) b->Return();
    return rc;
}

//======================================  Wait for a message to arrive
error_type 
TransportTCP::waitMsg(wtime_type timeout) {
    if (!mSocket)   return NotOpen; 
    if (isClient()) {
      return mSocket->wait(timeout); // wait for a client request.
    }
    
    if (!mCSocket) {
        //------------------------------  No connection - wait for one
        error_type rc = mSocket->waitConnect(timeout, &mCSocket);
	if (rc) return rc;
	if (is_async()) mCSocket->async();
	//return rc;
    }
    
    //----------------------------------  Wait for a client request.
    return mCSocket->wait(timeout);
}
